import os

import numpy as np

import iris
import iris.cube
import iris.analysis

import warnings
import datetime

import numba
from numba import vectorize
from numba import guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

import diagonals

__all__ = ['compute']


def compute(basins, vo):
    area = {}
    for basin in basins:
        area[basin] = _compute_area(e1v, e3v, basins[basin])

    if diagonals.CONFIG.use_gpu:
        moc = _compute_moc_gpu(
            vo, area
            )
    else:
        moc = _compute_moc_cpu(
            vo, area
        )
    return moc


def _compute_moc_cpu(vo, area):
    moc = []
    for basin in area:
        moc_basin = np.sum(_multiply_vo_basin(vo, area[basin]), axis=3)
        _vertical_cumsum(moc_basin, moc_basin)
        moc.append(moc_basin)
    return moc


def _compute_moc_gpu(vo, area):
    times = vo.shape[0]
    levels = vo.shape[1]
    lats = vo.shape[2]

    block = (128, 1, 1)
    grid_size = (lats // block[0]) + 1
    grid3D = (grid_size, levels, times)
    gpu_vo = cuda.to_device(vo.astype(np.float32))
    gpu_moc = cuda.device_array((times, levels, lats), dtype=np.float32)
    grid2D = (grid_size, times)

    moc = []
    for basin in area:
        gpu_area = cuda.to_device(area[basin].astype(np.float32))
        _horizontal_integral[grid3D, block](gpu_vo, gpu_area, gpu_moc)
        _vertical_cumsum_gpu[grid2D, block](gpu_moc)
        moc.append(gpu_moc.copy_to_host())

    del gpu_area, gpu_moc, gpu_vo

    return moc


@vectorize(['float32(float32, float32, float32)'], target='cpu')
def _compute_area(e1v, e3v, basin):
    return - e1v * e3v * basin / 1e6


@vectorize(['float32(float32, float32)'], target='cpu')
def _multiply_vo_basin(vo, basin):
    return vo*basin


@guvectorize([(float32[:, :, :], float32[:, :, :])], '(t, l, j)->(t, l, j)',
             target='cpu')
def _vertical_cumsum(moc, out):
    for lev in reversed(range(moc.shape[1]-1)):
        moc[:, lev, :] = moc[:, lev, :] + moc[:, lev+1, :]


@cuda.jit()
def _vertical_cumsum_gpu(moc):
    j, t = cuda.grid(2)
    if(j >= moc.shape[2]):
        return
    for lev in range(moc.shape[1]-2, -1, -1):
        moc[t, lev, j] = moc[t, lev, j] + moc[t, lev+1, j]


@cuda.jit()
def horizontal_integral(vo, area, moc):
    j, lev, t = cuda.grid(3)
    moc[t, lev, j] = 0.0
    temp = 0.0
    if(j >= vo.shape[2]):
        return
    for i in range(vo.shape[3]):
        temp += vo[t, lev, j, i] * area[0, lev, j, i]
    moc[t, lev, j] = temp